// SPDX-License-Identifier: BSD-3-Clause

#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdatomic.h>

#include "consumer.h"
#include "ring_buffer.h"
#include "packet.h"
#include "utils.h"

pthread_mutex_t consumer_dequeue_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t consumer_write_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_barrier_t write_barrier;
writer_data_t *writers;

// common variables
int out_file_fd;
int num_of_threads;

void add_packet(char *packet, int writer_id)
{
	int rez = process_packet((so_packet_t *)packet);
	unsigned long hash = packet_hash((so_packet_t *)packet);
	unsigned long timestamp = ((so_packet_t *)packet)->hdr.timestamp;

	// print the packet log
	writers[writer_id].rez = rez;
	writers[writer_id].hash = hash;
	writers[writer_id].timestamp = timestamp;

	writers[writer_id].status = 1;
}

int comparator(const void *a, const void *b)
{
	writer_data_t *writer_a = (writer_data_t *)a;
	writer_data_t *writer_b = (writer_data_t *)b;

	return writer_a->timestamp - writer_b->timestamp;
}

void sort_packets(void)
{
	qsort(writers, num_of_threads, sizeof(writer_data_t), comparator);
}

void empty_packet_array(void)
{
	memset(writers, 0, num_of_threads * sizeof(writer_data_t));
}

int everyone_is_finished(void)
{
	for (int i = 0; i < num_of_threads; i++) {
		if (!writers[i].status)
			return 0;
	}

	return 1;
}

#define SS "%s %016lx %lu\n"
#define WW writers
#define SPRT snprintf
void print_results_to_file(void)
{
	for (int i = 0; i < num_of_threads; i++) {
		char otbf[PKT_SZ];
		int len = SPRT(otbf, 256, SS, RES_TO_STR(WW[i].rez), WW[i].hash, WW[i].timestamp);

		write(out_file_fd, otbf, len);
	}
}


void *consumer_thread(void *args)
{
	/* TODO: implement consumer thread */
	so_consumer_ctx_t *ctx = (so_consumer_ctx_t *)args;

	char *packet = calloc(1, PKT_SZ);
	char *empty_packet = calloc(1, PKT_SZ);

	int to_be_continued = 1;

	while (to_be_continued) {
		pthread_mutex_lock(&consumer_dequeue_lock);

		if ((ctx->producer_rb->len > 0) ||
			(ctx->producer_rb->producer_is_alive && ctx->producer_rb->len == 0))
			ring_buffer_dequeue(ctx->producer_rb, packet, PKT_SZ);

		if ((!ctx->producer_rb->producer_is_alive && ctx->producer_rb->len == 0) ||
			(memcmp(packet, empty_packet, PKT_SZ) == 0)) {
			// pthread_mutex_unlock(&ctx->dequeue_access_mutex);
			// printf("cazul ala fortza\tthread: %d\n", ctx->thread_number);
			// break;
			to_be_continued = 0;
		}

		pthread_mutex_unlock(&consumer_dequeue_lock);

		if (memcmp(packet, empty_packet, PKT_SZ) == 0) {
			// adauga 0 la pacjetul lui
			// add_packet(packet, ctx->thread_number);
			// mark the empty packet as filled
			// writers[ctx->thread_number].status = 1;
			break;
		}

		// add packet to be sorted
		add_packet(packet, ctx->thread_number);

		// wait for everyone to finish
		// printf("thread: %d ready!\n", ctx->thread_number);
		pthread_barrier_wait(&write_barrier);
		// printf("barier broken\n");

		// critical operation of writing to the out file
		pthread_mutex_lock(&consumer_write_lock);
		if (everyone_is_finished()) {
			sort_packets();
			print_results_to_file();
			empty_packet_array();
		}
		pthread_mutex_unlock(&consumer_write_lock);

		memset(packet, 0, PKT_SZ);
	}

	free(packet);
	free(empty_packet);
	free(ctx);
	return NULL;
}

int create_consumers(pthread_t *tids,
					 int num_consumers,
					 struct so_ring_buffer_t *rb,
					 const char *out_filename)
{
	pthread_barrier_init(&write_barrier, NULL, num_consumers);

	// array of num_consumers packets that will be sorted
	writers = calloc(num_consumers, sizeof(writer_data_t));

	for (int i = 0; i < num_consumers; i++) {
		/* TODO: Launch consumer threads */
		so_consumer_ctx_t *ctx = calloc(1, sizeof(so_consumer_ctx_t));

		ctx->producer_rb = rb;
		ctx->out_file_name = out_filename;
		ctx->thread_number = i;
		pthread_create(&tids[i], NULL, consumer_thread, (void *)ctx);
	}

	return num_consumers;
}

void destroy_remainings(void)
{
	pthread_mutex_destroy(&consumer_dequeue_lock);
	pthread_mutex_destroy(&consumer_write_lock);
	pthread_barrier_destroy(&write_barrier);

	free(writers);
}

void set_common_variables(int outus_fd, int thrd_cnt)
{
	out_file_fd = outus_fd;
	num_of_threads = thrd_cnt;
}
