// SPDX-License-Identifier: BSD-3-Clause
#include <stdio.h>
#include "ring_buffer.h"

#define PKT_SIZE 256

int is_initialised;
pthread_mutex_t init_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t producer_started = PTHREAD_COND_INITIALIZER;

/*
 * @brief Initialises the ring buffer
 *
 * If ring buffer is not initialised it gets initialised with most values set to 0.
 *
 * @note Function is run only by one thread
 * @note Operation is not thread safe, implementation relies one thread calling this func!
 *
 * @returns 0 regardless of success rate
 */
int ring_buffer_init(so_ring_buffer_t *ring, size_t cap)
{
	/* TODO: implement ring_buffer_init */
	pthread_mutex_init(&ring->mutex, NULL);
	sem_init(&ring->rb_full, 0, 0);
	sem_init(&ring->rb_empty, 0, cap);
	pthread_cond_init(&ring->start_consumers, NULL);

	ring->producer_is_alive = 1;
	ring->cap = cap;
	ring->len = 0;
	ring->read_pos = 0;
	ring->write_pos = 0;
	ring->data = calloc(cap, PKT_SIZE);

	return 0;
}

/*
 * @brief Adds data to the ring buffer
 *
 * Producer creates packets that are inserted into the ring buffer.
 *
 * @note Code assumes that ring buffer always exists when this function is called
 * @returns -1 in case of error
 */
ssize_t ring_buffer_enqueue(so_ring_buffer_t *ring, void *data, size_t size)
{
	int all_good = -1;
	/* TODO: implement ring_buffer_enqueue */
	sem_wait(&ring->rb_empty);

	pthread_mutex_lock(&ring->mutex);
	// printf("inside critical section of enqueue\n");
	memcpy(ring->data + ring->write_pos * PKT_SIZE, data, size);
	if (!memcmp(ring->data + ring->write_pos, data, size))
		all_good = 0;

	ring->len++;
	ring->write_pos++;
	if (ring->write_pos >= ring->cap)
		ring->write_pos = 0;

	sem_post(&ring->rb_full);

	pthread_mutex_unlock(&ring->mutex);
	return all_good;
}


//HOUSTON, WE HAVE PROBLEMS HERE!!!!!
/*
 * @brief Removes data to the ring buffer
 *
 * Consumer removes packets from the ring buffer into the data.
 *
 * @param data destination
 * @param ring source
 *
 * @note Code assumes that ring buffer always exists when this function is called
 * @returns -1 in case of error
 */
ssize_t ring_buffer_dequeue(so_ring_buffer_t *ring, void *data, size_t size)
{
	int all_good = -1;
	/* TODO: Implement ring_buffer_dequeue */
	sem_wait(&ring->rb_full);

	pthread_mutex_lock(&ring->mutex);
	// printf("inside critical section of dequeue\n");
	if (ring->len == 0) {
		pthread_mutex_unlock(&ring->mutex);
		printf("exiting weird case\n");
		return -1;
	}

	memcpy(data, ring->data + ring->read_pos * PKT_SIZE, size);
	if (!memcmp(ring->data + ring->read_pos, data, size))
		all_good = 0;
	memset(ring->data + (ring->read_pos) * PKT_SIZE, 0, size);

	ring->len--;

	ring->read_pos++;
	if (ring->read_pos >= ring->cap)
		ring->read_pos = 0;

	sem_post(&ring->rb_empty);

	pthread_mutex_unlock(&ring->mutex);
	// printf("exiting dequeue\n");
	return all_good;
}

/*
 * @brief Destroy the ring buffer
 *
 * Dynamically allocated data is freed and syncronistaion primitives are destroyed.
 *
 * @note Operation is not thread safe,implementation assumes
 * one thread destroyes the ring buffer
 **/
void ring_buffer_destroy(so_ring_buffer_t *ring)
{
	/* TODO: Implement ring_buffer_destroy */
	pthread_mutex_destroy(&ring->mutex);
	sem_destroy(&ring->rb_full);
	sem_destroy(&ring->rb_empty);
	pthread_cond_destroy(&ring->start_consumers);

	free(ring->data);
}

/*
 * @brief Stops a thread
 *
 * Makes a thread drop its current execution early.
 *
 * Function is ideal for when the ring buffer is terminated
 * but there still are threads runing something, i guess
 */
void ring_buffer_stop(so_ring_buffer_t *ring)
{
	/* TODO: Implement ring_buffer_stop */
	pthread_mutex_lock(&ring->mutex);
	ring->producer_is_alive = 0;
	sem_post(&ring->rb_full);
	// printf("producer stopped\n");
	pthread_mutex_unlock(&ring->mutex);
}
